# Classes to model a HDL design hierarchy
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Functionality to represent and operate on a HDL code project
"""
from typing import Optional, Union
from pathlib import Path
import logging
from collections import OrderedDict
from vunit.hashing import hash_string
from vunit.dependency_graph import DependencyGraph, CircularDependencyException
from vunit.vhdl_parser import VHDLParser
from vunit.parsing.verilog.parser import VerilogParser
from vunit.exceptions import CompileError
from vunit import ostools
from vunit.source_file import (
    VERILOG_FILE_TYPES,
    SourceFile,
    VerilogSourceFile,
    VHDLSourceFile,
)
from vunit.vhdl_standard import VHDL, VHDLStandard
from vunit.library import Library

LOGGER = logging.getLogger(__name__)


class Project(object):  # pylint: disable=too-many-instance-attributes
    """
    The representation of a HDL code project.
    Compute lists of source files to recompile based on file contents,
    timestamps and depenencies derived from the design hierarchy.
    """

    def __init__(self, depend_on_package_body=False, database=None):
        """
        depend_on_package_body - Package users depend also on package body
        """
        self._database = database
        self._vhdl_parser = VHDLParser(database=self._database)
        self._verilog_parser = VerilogParser(database=self._database)
        self._libraries = OrderedDict()
        # Mapping between library lower case name and real library name
        self._lower_library_names_dict = {}
        self._source_files_in_order = []
        self._manual_dependencies = []
        self._depend_on_package_body = depend_on_package_body
        self._builtin_libraries = set(["ieee", "std"])

    def _validate_new_library_name(self, library_name):
        """
        Check that the library_name is valid or raise RuntimeError
        """
        if library_name == "work":
            LOGGER.error(
                "Cannot add library named work. work is a reference to the current library. "
                "http://www.sigasi.com/content/work-not-vhdl-library"
            )
            raise RuntimeError("Illegal library name 'work'")

        if library_name in self._libraries:
            raise ValueError(f"Library {library_name!s} already exists")

        lower_name = library_name.lower()
        if lower_name in self._lower_library_names_dict:
            raise RuntimeError(
                f"Library name {library_name!r} not case-insensitive unique. "
                f"Library name {self._lower_library_names_dict[lower_name]!r} previously defined"
            )

    def add_builtin_library(self, logical_name):
        """
        Add a builtin library name that does not give missing dependency warnings
        """
        self._builtin_libraries.add(logical_name)

    def add_library(
        self,
        logical_name,
        directory: Union[str, Path],
        vhdl_standard: VHDLStandard = VHDL.STD_2008,
        is_external=False,
    ):
        """
        Add library to project with logical_name located or to be located in directory
        is_external -- Library is assumed to a black-box
        """
        self._validate_new_library_name(logical_name)

        dpath = Path(directory)
        dstr = str(directory)

        if is_external:
            if not dpath.exists():
                raise ValueError(f"External library {dstr!r} does not exist")

            if not dpath.is_dir():
                raise ValueError(f"External library must be a directory. Got {dstr!r}")

        library = Library(logical_name, dstr, vhdl_standard, is_external=is_external)
        LOGGER.debug("Adding library %s with path %s", logical_name, dstr)

        self._libraries[logical_name] = library
        self._lower_library_names_dict[logical_name.lower()] = library.name

    def add_source_file(  # pylint: disable=too-many-arguments
        self,
        file_name,
        library_name,
        file_type="vhdl",
        include_dirs=None,
        defines=None,
        vhdl_standard: Optional[VHDLStandard] = None,
        no_parse=False,
    ):
        """
        Add a file_name as a source file in library_name with file_type

        :param no_parse: Do not parse file contents
        """
        fname = file_name if isinstance(file_name, Path) else Path(file_name)
        if not fname.exists():
            raise ValueError(f"File {str(fname)!r} does not exist")

        LOGGER.debug("Adding source file %s to library %s", str(fname), library_name)
        library = self._libraries[library_name]

        if file_type == "vhdl":
            assert include_dirs is None
            source_file: SourceFile = VHDLSourceFile(
                fname,
                library,
                vhdl_parser=self._vhdl_parser,
                database=self._database,
                vhdl_standard=library.vhdl_standard if vhdl_standard is None else vhdl_standard,
                no_parse=no_parse,
            )
        elif file_type in VERILOG_FILE_TYPES:
            source_file = VerilogSourceFile(
                file_type,
                fname,
                library,
                verilog_parser=self._verilog_parser,
                database=self._database,
                include_dirs=include_dirs,
                defines=defines,
                no_parse=no_parse,
            )
        else:
            raise ValueError(file_type)

        old_source_file = library.add_source_file(source_file)
        if id(source_file) == id(old_source_file):
            self._source_files_in_order.append(source_file)

        return old_source_file

    def add_manual_dependency(self, source_file, depends_on):
        """
        Add manual dependency where 'source_file' depends_on 'depends_on'
        """
        self._manual_dependencies.append((source_file, depends_on))

    @staticmethod
    def _find_primary_secondary_design_unit_dependencies(source_file):
        """
        Iterate over dependencies between the primary design units of the source_file
        and their secondary design units
        """
        library = source_file.library

        for unit in source_file.design_units:
            if unit.is_primary:
                continue

            try:
                primary_unit = library.primary_design_units[unit.primary_design_unit]
            except KeyError:
                LOGGER.warning(
                    "%s: failed to find a primary design unit '%s' in library '%s'",
                    source_file.name,
                    unit.primary_design_unit,
                    library.name,
                )
            else:
                yield primary_unit.source_file

    def _find_vhdl_library_reference(self, library_name):
        """
        Find a VHDL library reference that is case insensitive or raise KeyError
        """
        real_library_name = self._lower_library_names_dict[library_name.lower()]
        return self._libraries[real_library_name]

    def _find_other_vhdl_design_unit_dependencies(  # pylint: disable=too-many-branches
        self, source_file, depend_on_package_body, implementation_dependencies
    ):
        """
        Iterate over the dependencies on other design unit of the source_file
        """
        for ref in source_file.dependencies:
            try:
                library = self._find_vhdl_library_reference(ref.library)
            except KeyError:
                if ref.library not in self._builtin_libraries:

                    LOGGER.warning("%s: failed to find library '%s'", source_file.name, ref.library)
                continue

            if ref.is_entity_reference() and ref.design_unit in library.modules:
                # Is a verilog module instantiation
                yield library.modules[ref.design_unit].source_file
                continue

            try:
                primary_unit = library.primary_design_units[ref.design_unit]
            except KeyError:
                if not library.is_external:
                    LOGGER.warning(
                        "%s: failed to find a primary design unit '%s' in library '%s'",
                        source_file.name,
                        ref.design_unit,
                        library.name,
                    )
                continue
            else:
                yield primary_unit.source_file

            if ref.is_entity_reference():
                if ref.reference_all_names_within():
                    # Reference all architectures,
                    # We make configuration declarations implicitly reference all architectures
                    names = primary_unit.architecture_names.keys()
                elif ref.name_within is None and implementation_dependencies:
                    # For implementation dependencies we add a dependency to all architectures
                    names = primary_unit.architecture_names.keys()
                else:
                    names = [ref.name_within]

                for name in names:
                    if name is None:
                        # Was not a reference to a specific architecture
                        continue

                    if name in primary_unit.architecture_names:
                        file_name = primary_unit.architecture_names[name]
                        yield library.get_source_file(file_name)
                    else:
                        LOGGER.warning(
                            "%s: failed to find architecture '%s' of entity '%s.%s'",
                            source_file.name,
                            name,
                            library.name,
                            primary_unit.name,
                        )

            elif ref.is_package_reference() and depend_on_package_body:
                try:
                    yield library.get_package_body(primary_unit.name).source_file
                except KeyError:
                    # There was no package body, which is legal in VHDL
                    pass

    def _find_verilog_package_dependencies(self, source_file):
        """
        Find dependencies from import of verilog packages
        """
        for package_name in source_file.package_dependencies:
            for library in self._libraries.values():
                try:
                    design_unit = library.verilog_packages[package_name]
                    yield design_unit.source_file
                except KeyError:
                    pass

    def _find_verilog_module_dependencies(self, source_file):
        """
        Find dependencies from instantiation of verilog modules
        """
        for module_name in source_file.module_dependencies:
            if module_name in source_file.library.modules:
                design_unit = source_file.library.modules[module_name]
                yield design_unit.source_file
            else:
                for library in self._libraries.values():
                    try:
                        design_unit = library.modules[module_name]
                        yield design_unit.source_file
                    except KeyError:
                        pass

    @staticmethod
    def _find_component_design_unit_dependencies(source_file):
        """
        Iterate over the dependencies on other design units of the source_file
        that are the result of component instantiations
        """
        for unit_name in source_file.depending_components:
            found_component_match = False

            try:
                primary_unit = source_file.library.primary_design_units[unit_name]
                yield primary_unit.source_file

                for file_name in primary_unit.architecture_names.values():
                    yield source_file.library.get_source_file(file_name)
            except KeyError:
                pass
            else:
                found_component_match = True

            try:
                module = source_file.library.modules[unit_name]
            except KeyError:
                pass
            else:
                found_component_match = True
                yield module.source_file

            if not found_component_match:
                LOGGER.debug(
                    "failed to find a matching entity/module for component '%s' ",
                    unit_name,
                )

    def create_dependency_graph(self, implementation_dependencies=False):
        """
        Create a DependencyGraph object of the HDL code project
        """

        def add_dependency(start, end):
            """
            Utility to add dependency
            """
            if start.name == end.name:
                return

            is_new = dependency_graph.add_dependency(start, end)

            if is_new:
                LOGGER.debug("Adding dependency: %s depends on %s", end.name, start.name)

        def add_dependencies(dependency_function, files):
            """
            Utility to add all dependencies returned by a dependency_function
            returning an iterator of dependencies
            """
            for source_file in files:
                for dependency in dependency_function(source_file):
                    add_dependency(dependency, source_file)

        dependency_graph = DependencyGraph()
        for source_file in self._source_files_in_order:
            dependency_graph.add_node(source_file)

        vhdl_files = [source_file for source_file in self._source_files_in_order if source_file.file_type == "vhdl"]

        depend_on_package_bodies = self._depend_on_package_body or implementation_dependencies
        add_dependencies(
            lambda source_file: self._find_other_vhdl_design_unit_dependencies(
                source_file, depend_on_package_bodies, implementation_dependencies
            ),
            vhdl_files,
        )
        add_dependencies(self._find_primary_secondary_design_unit_dependencies, vhdl_files)

        verilog_files = [
            source_file for source_file in self._source_files_in_order if source_file.file_type in VERILOG_FILE_TYPES
        ]

        add_dependencies(self._find_verilog_package_dependencies, verilog_files)
        add_dependencies(self._find_verilog_module_dependencies, verilog_files)

        if implementation_dependencies:
            add_dependencies(self._find_component_design_unit_dependencies, vhdl_files)

        for source_file, depends_on in self._manual_dependencies:
            add_dependency(depends_on, source_file)

        return dependency_graph

    @staticmethod
    def _handle_circular_dependency(exception):
        """
        Pretty print circular dependency to error log
        """
        LOGGER.error(
            "Found circular dependency:\n%s",
            " ->\n".join(source_file.name for source_file in exception.path),
        )

    def _get_compile_timestamps(self, files):
        """
        Return a dictionary of mapping file to the timestamp when it
        was compiled or None if it was not compiled
        """
        # Cache timestamps to avoid duplicate file operations
        timestamps = {}
        for source_file in files:
            hash_file_name = self._hash_file_name_of(source_file)
            if not ostools.file_exists(hash_file_name):
                timestamps[source_file] = None
            else:
                timestamps[source_file] = ostools.get_modification_time(hash_file_name)
        return timestamps

    def get_files_in_compile_order(self, incremental=True, dependency_graph=None, files=None):
        """
        Get a list of all files in compile order
        param: incremental: Only return files that need recompile if True
        param: files: provide a list of files that shall be sorted, if None all files are used
        """
        if dependency_graph is None:
            dependency_graph = self.create_dependency_graph()

        files_to_recompile = self._get_files_to_recompile(
            files or self.get_source_files_in_order(), dependency_graph, incremental
        )
        return self._get_affected_files_in_compile_order(files_to_recompile, dependency_graph.get_dependent)

    def _get_files_to_recompile(self, files, dependency_graph, incremental):
        """
        Analyse a given set of SourceFile according to the compile timestamps
        and return the set that has to be recompiled.
        param: files: a list of type SourceFile
        param: dependency_graph: The DependencyGraph object to be used
        """
        timestamps = self._get_compile_timestamps(files)
        result_list = []
        for source_file in files:
            if (not incremental) or self._needs_recompile(dependency_graph, source_file, timestamps):
                result_list.append(source_file)
        return result_list

    def get_dependencies_in_compile_order(self, target_files=None, implementation_dependencies=False):
        """
        Get a list of dependencies of target files including the
        target files.
        :param target_files: A list of SourceFiles
        """

        if target_files is None:
            target_files = self._source_files_in_order

        dependency_graph = self.create_dependency_graph(implementation_dependencies)
        return self._get_affected_files_in_compile_order(set(target_files), dependency_graph.get_dependencies)

    def _get_affected_files_in_compile_order(self, target_files, get_depend_func):
        """
        Returns the affected files in compile order given a list of target files and a dependencie function
        :param target_files: The files to compile
        :param get_depend_func: one of DependencyGraph [get_dependencies, get_dependent, get_direct_dependencies]
        """
        affected_files = self._get_affected_files(target_files, get_depend_func)
        return self._get_compile_order(affected_files, get_depend_func.__self__)

    def get_minimal_file_set_in_compile_order(self, target_files=None):
        """
        Get the minimal set of files to be compiled for a list of target files of type SourceFile
        param: target_files: List of type SourceFile, if the paramater is None all files are used
        """
        ###
        # First get all files that are required to fullfill the dependencies for the target files
        dependency_graph = self.create_dependency_graph(True)
        dependency_files = self._get_affected_files(
            target_files or self.get_source_files_in_order(),
            dependency_graph.get_dependencies,
        )

        ###
        # Now the file set is known, but it has to be evaluated which files
        # realy have to be compiled according to their timestamp.
        max_file_set_to_be_compiled = self.get_files_in_compile_order(incremental=True, files=dependency_files)

        # get_files_in_compile_order returns more files than actually are in the
        # list of dependent files. So the list is filtered for only the files
        # that are required
        min_file_set_to_be_compiled = [f for f in max_file_set_to_be_compiled if f in dependency_files]
        return min_file_set_to_be_compiled

    def _get_affected_files(self, target_files, get_depend_func):
        """
        Get affected files given a  list of type SourceFile, if the list is None
        all files are taken into account
        :param target_files: An initial list of type SourceFile
        :param get_depend_func: One of either [get_dependent, get_dependencies, get_direct_dependencies]
        of an object dependency_graph of type DependencyGraph
        """
        try:
            return get_depend_func(target_files)
        except CircularDependencyException as exc:
            self._handle_circular_dependency(exc)
            raise CompileError from exc

    def _get_compile_order(self, files, dependency_graph):
        """
        Returns a sorted list of type SourceFile using the given dependency graph
        param: dependency_graph: The DependencyGraph object
        """
        try:
            compile_order = dependency_graph.toposort()
        except CircularDependencyException as exc:
            self._handle_circular_dependency(exc)
            raise CompileError from exc

        def comparison_key(source_file):
            return compile_order.index(source_file)

        return sorted(files, key=comparison_key)

    def get_source_files_in_order(self):
        """
        Get a list of source files in the order they were added to the project
        """
        return list(self._source_files_in_order)

    def get_libraries(self):
        return self._libraries.values()

    def get_library(self, library_name):
        return self._libraries[library_name]

    def has_library(self, library_name):
        return library_name in self._libraries

    def _needs_recompile(self, dependency_graph, source_file, timestamps):
        """
        Returns True if the source_file needs to be recompiled
        given the dependency_graph, the file contents and the last modification time
        """
        timestamp = timestamps[source_file]

        content_hash_file_name = self._hash_file_name_of(source_file)
        if timestamp is None:
            LOGGER.debug(
                "%s has no vunit_hash file at %s and must be recompiled",
                source_file.name,
                content_hash_file_name,
            )
            return True

        old_content_hash = ostools.read_file(content_hash_file_name)
        if old_content_hash != source_file.content_hash:
            LOGGER.debug(
                "%s has different hash than last time and must be recompiled",
                source_file.name,
            )
            return True

        for other_file in dependency_graph.get_direct_dependencies(source_file):
            other_timestamp = timestamps[other_file]

            if other_timestamp is None:
                # Other file has not been compiled and will trigger recompile of this file
                continue

            if other_timestamp > timestamp:
                LOGGER.debug(
                    "%s has dependency compiled earlier and must be recompiled",
                    source_file.name,
                )
                return True

        LOGGER.debug("%s has same hash file and must not be recompiled", source_file.name)

        return False

    def _hash_file_name_of(self, source_file):
        """
        Returns the name of the hash file associated with the source_file
        """
        library = self.get_library(source_file.library.name)
        prefix = hash_string(str(Path(source_file.name).parent))
        return str(Path(library.directory) / prefix / Path(source_file.name).name / ".vunit_hash")

    def update(self, source_file):
        """
        Mark that source_file has been recompiled, triggers a re-write of the hash file
        to update the timestamp
        """
        new_content_hash = source_file.content_hash
        ostools.write_file(self._hash_file_name_of(source_file), new_content_hash)
        LOGGER.debug("Wrote %s content_hash=%s", source_file.name, new_content_hash)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
PyPI setup script
"""

import os
import sys
from pathlib import Path
from logging import warning
from setuptools import setup

# Ensure that the source tree is on the sys path
sys.path.insert(0, str(Path(__file__).parent.resolve()))

from vunit.about import version, doc  # pylint: disable=wrong-import-position
from vunit.builtins import osvvm_is_installed  # pylint: disable=wrong-import-position


def find_all_files(directory, endings=None):
    """
    Recursively find all files within directory
    """
    result = []
    for root, _, filenames in os.walk(directory):
        for filename in filenames:
            ending = os.path.splitext(filename)[-1]
            if endings is None or ending in endings:
                result.append(str(Path(root) / filename))
    return result


DATA_FILES = []
DATA_FILES += find_all_files("vunit", endings=[".tcl"])
DATA_FILES += find_all_files(str(Path("vunit") / "vhdl"))
DATA_FILES += find_all_files(str(Path("vunit") / "verilog"), endings=[".v", ".sv", ".svh"])
DATA_FILES = [os.path.relpath(file_name, "vunit") for file_name in DATA_FILES]

setup(
    name="vunit_hdl",
    version=version(),
    packages=[
        "tests",
        "tests.lint",
        "tests.unit",
        "tests.acceptance",
        "vunit",
        "vunit.com",
        "vunit.parsing",
        "vunit.parsing.verilog",
        "vunit.sim_if",
        "vunit.test",
        "vunit.ui",
        "vunit.vivado",
    ],
    package_data={"vunit": DATA_FILES},
    zip_safe=False,
    url="https://github.com/VUnit/vunit",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development :: Testing",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ],
    python_requires=">=3.6",
    install_requires=["colorama"],
    requires=["colorama"],
    license="Mozilla Public License 2.0 (MPL 2.0)",
    author="Lars Asplund",
    author_email="lars.anders.asplund@gmail.com",
    description="VUnit is an open source unit testing framework for VHDL/SystemVerilog.",
    long_description=doc(),
)

if not osvvm_is_installed():
    warning(
        """
Found no OSVVM VHDL files. If you're installing from a Git repository and plan to use VUnit's integration
of OSVVM you should run

git submodule update --init --recursive

in your VUnit repository before running setup.py."""
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Preprocessing of VHDL files to add file_name and line_num arguments to function calls
to enable better messages
"""


import re


class LocationPreprocessor(object):
    """
    Preprocessing of VHDL files to add file_name and line_num
    arguments to calls of known function and procedures
    """

    def __init__(self):
        self._subprograms_with_arguments = [
            "log",
            "verbose_high2",
            "verbose_high1",
            "verbose",
            "verbose_low1",
            "verbose_low2",
            "debug_high2",
            "debug_high1",
            "debug",
            "debug_low1",
            "debug_low2",
            "info_high2",
            "info_high1",
            "info",
            "info_low1",
            "info_low2",
            "warning_high2",
            "warning_high1",
            "warning",
            "warning_low1",
            "warning_low2",
            "error_high2",
            "error_high1",
            "error",
            "error_low1",
            "error_low2",
            "failure_high2",
            "failure_high1",
            "failure",
            "failure_low1",
            "failure_low2",
            "check",
            "check_failed",
            "check_true",
            "check_false",
            "check_implication",
            "check_stable",
            "check_equal",
            "check_not_unknown",
            "check_zero_one_hot",
            "check_one_hot",
            "check_next",
            "check_sequence",
            "check_relation",
            "lock_entry",
            "lock_exit",
            "unlock_entry",
            "unlock_exit",
        ]
        self._subprograms_without_arguments = []

    def add_subprogram(self, subprogram):
        """
        Add a subprogram name to the list of known names to preprocess
        """
        self._subprograms_without_arguments.append(subprogram)
        self._subprograms_with_arguments.append(subprogram)

    def remove_subprogram(self, subprogram):
        """
        Remove a subprogram name from the list of known names to preprocess
        """
        if subprogram not in self._subprograms_without_arguments + self._subprograms_with_arguments:
            raise RuntimeError(f"Unable to remove unknown subprogram {subprogram!s}")

        if subprogram in self._subprograms_without_arguments:
            self._subprograms_without_arguments.remove(subprogram)

        if subprogram in self._subprograms_with_arguments:
            self._subprograms_with_arguments.remove(subprogram)

    @staticmethod
    def _find_closing_parenthesis(args):
        """
        Find the balanced closing parentesis

        @TODO duplicate with vhdl_parser.py
        """
        pattern = re.compile(r"\(|\)")
        balance = 0
        for match in pattern.finditer(args):
            balance += 1 if match.group() == "(" else -1
            if balance == 0:
                return match.start()
        return None

    _already_fixed_file_name_pattern = re.compile(r"file_name\s*=>", re.MULTILINE)
    _already_fixed_line_num_pattern = re.compile(r"line_num\s*=>", re.MULTILINE)
    _subprogram_declaration_start_backwards_pattern = re.compile(r"\s+(erudecorp|noitcnuf)")
    _assignment_pattern = re.compile(r"\s*(:=|<=)", re.MULTILINE)

    def run(self, code, file_name):
        """
        Return preprocessed code given file_name of original file
        """
        potential_subprogram_call_with_arguments_pattern = re.compile(
            r"[^a-zA-Z0-9_](?P<subprogram>" + "|".join(self._subprograms_with_arguments) + r")\s*(?P<args>\()",
            re.MULTILINE,
        )

        potential_subprogram_call_without_arguments_pattern = re.compile(
            r"[^a-zA-Z0-9_](?P<subprogram>" + "|".join(self._subprograms_without_arguments) + r")\s*;",
            re.MULTILINE,
        )

        matches = list(potential_subprogram_call_with_arguments_pattern.finditer(code))
        if self._subprograms_without_arguments:
            matches += list(potential_subprogram_call_without_arguments_pattern.finditer(code))
        matches.sort(key=lambda match: match.start("subprogram"), reverse=True)

        for match in matches:
            if self._subprogram_declaration_start_backwards_pattern.match(code[match.start() : 0 : -1]):
                continue
            file_name_association = ', file_name => "' + file_name + '"'
            line_num_association = ", line_num => " + str(1 + code[: match.start("subprogram")].count("\n"))
            if "args" in match.groupdict():
                closing_paranthesis_start = self._find_closing_parenthesis(code[match.start("args") :])

                if self._assignment_pattern.match(code[match.start("args") + closing_paranthesis_start + 1 :]):
                    continue

                args = code[match.start("args") : match.start("args") + closing_paranthesis_start]
                already_fixed_file_name = self._already_fixed_file_name_pattern.search(args) is not None
                already_fixed_line_num = self._already_fixed_line_num_pattern.search(args) is not None
                file_name_association = file_name_association if not already_fixed_file_name else ""
                line_num_association = line_num_association if not already_fixed_line_num else ""

                code = (
                    code[: match.start("args") + closing_paranthesis_start]
                    + line_num_association
                    + file_name_association
                    + code[match.start("args") + closing_paranthesis_start :]
                )
            else:
                code = (
                    code[: match.end("subprogram")]
                    + "("
                    + line_num_association[2:]
                    + file_name_association
                    + ")"
                    + code[match.end("subprogram") :]
                )
        return code



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the DependencyGraph
"""

import unittest
from vunit.dependency_graph import DependencyGraph, CircularDependencyException


class TestDependencyGraph(unittest.TestCase):
    """
    Test the DependencyGraph
    """

    def test_should_return_empty_compile_order_for_no_nodes(self):
        graph = DependencyGraph()
        self.assertEqual(graph.toposort(), [], "Should return empty list")

    def test_should_return_list_of_nodes_when_there_are_no_dependencies(self):
        nodes = ["a", "b", "c", "d"]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, [])
        result = graph.toposort()
        self.assertEqual(result.sort(), nodes.sort(), "Should return the node list in any order")

    def test_should_sort_in_topological_order_when_there_are_dependencies(self):
        nodes = ["a", "b", "c", "d", "e", "f"]
        dependencies = [("a", "b"), ("a", "c"), ("b", "d"), ("e", "f")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.toposort()
        self._check_result(result, dependencies)

    def test_should_raise_runtime_error_exception_on_self_dependency(self):
        nodes = ["a", "b", "c", "d"]
        dependencies = [("a", "b"), ("a", "c"), ("b", "d"), ("d", "d")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.toposort()
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ["d", "d"])
        else:
            self.fail("Exception not raised")

    def test_should_raise_runtime_error_exception_on_long_circular_dependency(self):
        nodes = ["a", "b", "c", "d"]
        dependencies = [("a", "b"), ("a", "c"), ("b", "d"), ("d", "a")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.toposort()
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ["a", "b", "d", "a"])
        else:
            self.fail("Exception not raised")

    def test_should_resort_after_additions(self):
        nodes = ["a", "b", "c", "d", "e", "f"]
        dependencies = [("a", "b"), ("a", "c"), ("b", "d"), ("e", "f")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        graph.toposort()
        dependencies = [("a", "b"), ("a", "c"), ("b", "d"), ("e", "f"), ("b", "g")]
        graph.add_node("g")
        graph.add_dependency("b", "g")
        result = graph.toposort()
        self._check_result(result, dependencies)

    def test_get_direct_dependencies_should_return_empty_set_when_no_dependendencies(
        self,
    ):
        nodes = ["a", "b", "c"]
        dependencies = []
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.get_direct_dependencies("b")
        self.assertTrue(isinstance(result, (set)))
        self.assertFalse(result)

    def test_get_direct_dependencies_should_return_dependendencies_set(self):
        nodes = ["a", "b", "c", "d"]
        dependencies = [("a", "b"), ("a", "c")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        result = graph.get_direct_dependencies("c")
        self.assertFalse("b" in result)
        self.assertTrue("a" in result)

    def test_get_dependent(self):
        nodes = ["a", "b", "c", "d", "e"]
        dependencies = [("a", "b"), ("b", "c"), ("c", "d"), ("e", "d")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        self.assertEqual(graph.get_dependent(set("a")), set(("a", "b", "c", "d")))
        self.assertEqual(graph.get_dependent(set("e")), set(("d", "e")))

    def test_get_dependencies(self):
        nodes = ["a", "b", "c", "d", "e"]
        dependencies = [("b", "a"), ("c", "b"), ("d", "c"), ("d", "e")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)
        self.assertEqual(graph.get_dependencies(set("a")), set(("a", "b", "c", "d")))
        self.assertEqual(graph.get_dependencies(set("e")), set(("d", "e")))

    def test_get_dependent_detects_circular_dependencies(self):
        nodes = ["a", "b", "c"]
        dependencies = [("a", "b"), ("b", "c"), ("c", "a")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.get_dependent(set("a"))
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ["a", "b", "c", "a"])
        else:
            self.fail("Exception not raised")

    def test_get_dependencies_detects_circular_dependencies(self):
        nodes = ["a", "b", "c"]
        dependencies = [("b", "a"), ("c", "b"), ("a", "c")]
        graph = DependencyGraph()
        self._add_nodes_and_dependencies(graph, nodes, dependencies)

        try:
            graph.get_dependencies(set("a"))
        except CircularDependencyException as exc:
            self.assertEqual(exc.path, ["a", "b", "c", "a"])
        else:
            self.fail("Exception not raised")

    def _check_result(self, result, dependencies):
        """
        Check that the resulting has an order such that
        dependent files are located after their dependencies
        """
        for dep1, dep2 in dependencies:
            self.assertTrue(
                result.index(dep1) < result.index(dep2),
                "%s is not before %s" % (dep1, dep2),
            )

    @staticmethod
    def _add_nodes_and_dependencies(graph, nodes, dependencies):
        """
        Add nodes and dependencies to the graph
        """
        for node in nodes:
            graph.add_node(node)
        for dep in dependencies:
            graph.add_dependency(dep[0], dep[1])




import json
from os.path import join, dirname
from vunit import VUnit
from base64 import b16encode


def json_load(p):
    return json.loads(open(p, "r").read())


def json_dump(o):
    return json.dumps(o, separators=(",", ":"))


def b16enc(s):
    return b16encode(bytes(s, 'utf-8')).decode('utf-8')


root = join(dirname(__file__), '..', '..')

vu = VUnit.from_argv()

lib = vu.add_library("JSON")
lib.add_source_files(join(root, "src", "*.vhdl"))
lib = vu.add_library("lib")
lib.add_source_files(join(root, "examples", "*_VUnit.vhdl"))

gen_cfg = json_load(join(root, 'data', 'Boards1.json'))
gen_str = json_dump(gen_cfg)

tb = lib.get_test_benches('*tb_boards*')[0]

tb.get_tests("stringified*")[0].set_generic("tb_cfg", gen_str)
tb.get_tests("b16encoded stringified*")[0].set_generic("tb_cfg", b16enc(gen_str))
tb.get_tests("JSON file*")[0].set_generic("tb_cfg", join(root, 'data', 'Boards0.json'))
tb.get_tests("b16encoded JSON file*")[0].set_generic("tb_cfg", b16enc(join(root, 'data', 'Boards0.json')))

vu.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
UI common functions
"""

from pathlib import Path
from glob import glob
from os import environ
from logging import getLogger
from typing import Optional, List
from ..sim_if import is_string_not_iterable
from ..vhdl_standard import VHDL, VHDLStandard

LOGGER = getLogger(__name__)

TEST_OUTPUT_PATH = "test_output"


def select_vhdl_standard(vhdl_standard: Optional[str] = None) -> VHDLStandard:
    """
    Select VHDL standard either from class initialization or according to environment variable VUNIT_VHDL_STANDARD
    """
    if vhdl_standard is not None:
        return VHDL.standard(vhdl_standard)

    try:
        return VHDL.standard(environ.get("VUNIT_VHDL_STANDARD", "2008"))
    except ValueError:
        LOGGER.error("Invalid standard set by VUNIT_VHDL_STANDARD environment variable")
        raise


def lower_generics(generics):
    """
    Convert all generics names to lower case to match internal representation.
    @TODO Maybe warn in case of conflict. VHDL forbids this though so the user will notice anyway.
    """
    return dict((name.lower(), value) for name, value in generics.items())


def check_not_empty(lst, allow_empty, error_msg):
    """
    Raise ValueError if the list is empty unless allow_empty is True
    Returns the list
    """
    if (not allow_empty) and (not lst):
        raise ValueError(error_msg + ". Use allow_empty=True to avoid exception.")
    return lst


def get_checked_file_names_from_globs(pattern, allow_empty):
    """
    Get file names from globs and check that exist
    """
    if is_string_not_iterable(pattern):
        patterns = [pattern]
    elif isinstance(pattern, Path):
        patterns = [str(pattern)]
    else:
        patterns = pattern

    file_names: List[str] = []
    for pattern_instance in patterns:
        new_file_names = glob(str(pattern_instance), recursive=True)
        check_not_empty(
            new_file_names,
            allow_empty,
            f"Pattern {pattern_instance!r} did not match any file",
        )
        file_names += new_file_names

    return file_names



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Contains type defining VHDL standards and operations on them
"""

from functools import total_ordering


@total_ordering
class VHDLStandard(object):
    """
    VHDL standard object which encapsulates knowledge about VHDL standards
    """

    _STANDARDS = {"1993", "2002", "2008", "2019"}

    def __init__(self, standard_name):
        if standard_name in self._STANDARDS:
            self._standard = standard_name
        else:
            for standard in self._STANDARDS:
                if standard.endswith(standard_name) and len(standard_name) == 2:
                    self._standard = standard
                    break
            else:
                raise ValueError(f"Unknown standard '{standard_name!s}'")

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._standard == other._standard  # pylint: disable=protected-access
        return False

    def __lt__(self, other):
        return int(self._standard) < int(other._standard)  # pylint: disable=protected-access

    def __str__(self):
        if self == VHDL.STD_1993:
            # For backwards compatibility due to legacy reasons
            return "93"
        return self._standard

    def __repr__(self):
        return f"VHDLStandard({self._standard!r})"

    def __hash__(self):
        return hash(self._standard)

    @property
    def supports_context(self):
        return self >= VHDL.STD_2008

    @property
    def and_later(self):
        """
        Return a set including this standard and all later standards
        """
        return {standard for standard in VHDL.STANDARDS if standard >= self}

    @property
    def and_earlier(self):
        """
        Return a set including this standard and all earlier standards
        """
        return {standard for standard in VHDL.STANDARDS if standard <= self}


class VHDL(object):
    """
    Just a namespace for standards
    """

    STD_1993 = VHDLStandard("1993")
    STD_2002 = VHDLStandard("2002")
    STD_2008 = VHDLStandard("2008")
    STD_2019 = VHDLStandard("2019")
    STANDARDS = [STD_1993, STD_2002, STD_2008, STD_2019]

    @staticmethod
    def standard(name):
        return VHDLStandard(name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
License header sanity check
"""

import unittest
from warnings import simplefilter, catch_warnings
from pathlib import Path
from os.path import commonprefix
from os import walk
import re
from vunit import ROOT as RSTR
from vunit.builtins import VHDL_PATH
from vunit import ostools
from vunit.about import license_text

ROOT = Path(RSTR)

RE_LICENSE_NOTICE = re.compile(
    r"(?P<comment_start>#|--|//) This Source Code Form is subject to the terms of the Mozilla Public" + "\n"
    r"(?P=comment_start) License, v\. 2\.0\. If a copy of the MPL was not distributed with this file," + "\n"
    r"(?P=comment_start) You can obtain one at http://mozilla\.org/MPL/2\.0/\." + "\n"
    r"(?P=comment_start)" + "\n"
    r"(?P=comment_start) Copyright \(c\) (?P<first_year>20\d\d)(-(?P<last_year>20\d\d))?, "
    + r"Lars Asplund lars\.anders\.asplund@gmail\.com"
)

RE_LOG_DATE = re.compile(r"Date:\s*(?P<year>20\d\d)-\d\d-\d\d")

FIRST_YEAR = 2014
LAST_YEAR = 2021


class TestLicense(unittest.TestCase):
    """
    Test that each file in the repository contains a valid license
    header with a correct year range.
    The correct year range is computed based on the commit history.
    """

    def test_that_a_valid_license_exists_in_source_files_and_that_global_licensing_information_is_correct(
        self,
    ):
        for file_name in find_licensed_files():
            code = ostools.read_file(file_name)
            self._check_license(code, file_name)
            if Path(file_name).suffix in (".vhd", ".vhdl", ".v", ".sv"):
                self._check_no_trailing_whitespace(code, file_name)

    def test_that_license_file_matches_vunit_license_text(self):
        with catch_warnings():
            simplefilter("ignore", category=DeprecationWarning)
            with (ROOT / "LICENSE.rst").open("rU") as lic:
                self.assertEqual(lic.read(), license_text())

    def _check_license(self, code, file_name):
        """
        Check that the license header of file_name is valid
        """

        match = RE_LICENSE_NOTICE.search(code)
        self.assertIsNotNone(match, "Failed to find license notice in %s" % file_name)
        self.assertEqual(
            int(match.group("first_year")),
            FIRST_YEAR,
            "Expected copyright year range to start with %d in %s" % (FIRST_YEAR, file_name),
        )
        self.assertEqual(
            int(match.group("last_year")),
            LAST_YEAR,
            "Expected copyright year range to end with %d in %s" % (LAST_YEAR, file_name),
        )

    @staticmethod
    def _check_no_trailing_whitespace(code, file_name):
        """
        Check that there is no trailing whitespace within the code
        """
        for idx, line in enumerate(code.splitlines()):
            sline = line.rstrip()
            if sline == line:
                continue
            line_prefix = "%i: " % (idx + 1)
            print("Trailing whitespace violation in %s" % file_name)
            print(line_prefix + line)
            for _ in range(len(line_prefix) + len(sline)):
                print(" ", end="")
            for _ in range(len(line) - len(sline)):
                print("~", end="")
            print()
            raise AssertionError("Line %i of %s contains trailing whitespace" % (idx + 1, file_name))


def fix_license(file_name):
    """
    Fix license notice in file
    """

    with open(file_name, "r") as fptr:
        text = fptr.read()

    replacement = "Copyright (c) %i-%i, Lars Asplund lars.anders.asplund@gmail.com" % (
        FIRST_YEAR,
        LAST_YEAR,
    )

    text = re.sub(
        r"Copyright \(c\) (?P<first_year>20\d\d)(-(?P<last_year>20\d\d))?, "
        r"Lars Asplund lars\.anders\.asplund@gmail\.com",
        replacement,
        text,
    )

    with open(file_name, "w") as fptr:
        fptr.write(text)


def find_licensed_files():
    """
    Return all licensed files
    """
    licensed_files = []
    for root, _, files in walk(RSTR):
        for file_name in files:
            if "preprocessed" in root:
                continue
            if "codecs" in root:
                continue
            if root == str(ROOT / "docs"):
                continue
            if str(ROOT / "venv") in root:
                continue
            if str(ROOT / ".tox") in root:
                continue
            if is_prefix_of(
                (VHDL_PATH / "osvvm").resolve(),
                (Path(root) / file_name).resolve(),
            ):
                continue
            if is_prefix_of(
                (VHDL_PATH / "JSON-for-VHDL").resolve(),
                (Path(root) / file_name).resolve(),
            ):
                continue
            if Path(file_name).suffix in (".vhd", ".vhdl", ".py", ".v", ".sv"):
                licensed_files.append(str(Path(root) / file_name))
    return licensed_files


def is_prefix_of(prefix: Path, of_path: Path):
    """
    Return True if 'prefix' is a prefix of 'of_path'
    """
    return commonprefix([str(prefix), str(of_path)]) == str(prefix)


def main():
    """
    Fix license notice in all licensed files
    """
    for file_name in find_licensed_files():
        fix_license(file_name)


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
MyPy check
"""

import unittest
import sys
from subprocess import check_call


class TestMyPy(unittest.TestCase):
    """
    Run MyPy static type analysis
    """

    @staticmethod
    def test_mypy():
        check_call([sys.executable, "-m", "mypy", "--namespace-packages", "vunit"])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Test bench, runner, etc.
"""
